import _regeneratorRuntime from 'babel-runtime/regenerator';
import _asyncToGenerator from 'babel-runtime/helpers/asyncToGenerator';
import _classCallCheck from 'babel-runtime/helpers/classCallCheck';
import _createClass from 'babel-runtime/helpers/createClass';
import randomToken from 'random-token';
import IdleQueue from 'custom-idle-queue';

import PouchDB from './pouch-db';
import * as util from './util';
import RxError from './rx-error';
import RxCollection from './rx-collection';
import RxSchema from './rx-schema';
import RxChangeEvent from './rx-change-event';
import Socket from './socket';
import overwritable from './overwritable';
import { runPluginHooks } from './hooks';

import { Subject } from 'rxjs';
import { filter } from 'rxjs/operators';

/**
 * stores the combinations
 * of used database-names with their adapters
 * so we can throw when the same database is created more then once
 * @type {Object<string, array>} map with {dbName -> array<adapters>}
 */
var USED_COMBINATIONS = {};

var DB_COUNT = 0;

export var RxDatabase = function () {
    function RxDatabase(name, adapter, password, multiInstance, options, pouchSettings) {
        _classCallCheck(this, RxDatabase);

        if (typeof name !== 'undefined') DB_COUNT++;
        this.name = name;
        this.adapter = adapter;
        this.password = password;
        this.multiInstance = multiInstance;
        this.options = options;
        this.pouchSettings = pouchSettings;
        this.idleQueue = new IdleQueue();
        this.token = randomToken(10);

        this._subs = [];
        this.destroyed = false;

        // cache for collection-objects
        this.collections = {};

        // rx
        this.subject = new Subject();
        this.observable$ = this.subject.asObservable().pipe(filter(function (cEvent) {
            return RxChangeEvent.isInstanceOf(cEvent);
        }));
    }

    RxDatabase.prototype.dangerousRemoveCollectionInfo = function () {
        var _ref = _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee() {
            var colPouch, docsRes;
            return _regeneratorRuntime.wrap(function _callee$(_context) {
                while (1) {
                    switch (_context.prev = _context.next) {
                        case 0:
                            colPouch = this._collectionsPouch;
                            _context.next = 3;
                            return colPouch.allDocs();

                        case 3:
                            docsRes = _context.sent;
                            _context.next = 6;
                            return Promise.all(docsRes.rows.map(function (row) {
                                return {
                                    _id: row.key,
                                    _rev: row.value.rev
                                };
                            }).map(function (doc) {
                                return colPouch.remove(doc._id, doc._rev);
                            }));

                        case 6:
                        case 'end':
                            return _context.stop();
                    }
                }
            }, _callee, this);
        }));

        function dangerousRemoveCollectionInfo() {
            return _ref.apply(this, arguments);
        }

        return dangerousRemoveCollectionInfo;
    }();

    /**
     * do the async things for this database
     */


    RxDatabase.prototype.prepare = function () {
        var _ref2 = _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee2() {
            var _this = this;

            var pwHashDoc;
            return _regeneratorRuntime.wrap(function _callee2$(_context2) {
                while (1) {
                    switch (_context2.prev = _context2.next) {
                        case 0:
                            if (!this.password) {
                                _context2.next = 22;
                                break;
                            }

                            _context2.next = 3;
                            return this.lockedRun(function () {
                                return _this._adminPouch.info();
                            });

                        case 3:
                            pwHashDoc = null;
                            _context2.prev = 4;
                            _context2.next = 7;
                            return this.lockedRun(function () {
                                return _this._adminPouch.get('_local/pwHash');
                            });

                        case 7:
                            pwHashDoc = _context2.sent;
                            _context2.next = 12;
                            break;

                        case 10:
                            _context2.prev = 10;
                            _context2.t0 = _context2['catch'](4);

                        case 12:
                            if (pwHashDoc) {
                                _context2.next = 20;
                                break;
                            }

                            _context2.prev = 13;
                            _context2.next = 16;
                            return this.lockedRun(function () {
                                return _this._adminPouch.put({
                                    _id: '_local/pwHash',
                                    value: util.hash(_this.password)
                                });
                            });

                        case 16:
                            _context2.next = 20;
                            break;

                        case 18:
                            _context2.prev = 18;
                            _context2.t1 = _context2['catch'](13);

                        case 20:
                            if (!(pwHashDoc && this.password && util.hash(this.password) !== pwHashDoc.value)) {
                                _context2.next = 22;
                                break;
                            }

                            throw RxError.newRxError('DB1', {
                                passwordHash: util.hash(this.password),
                                existingPasswordHash: pwHashDoc.value
                            });

                        case 22:
                            if (!this.multiInstance) {
                                _context2.next = 27;
                                break;
                            }

                            _context2.next = 25;
                            return Socket.create(this);

                        case 25:
                            this.socket = _context2.sent;


                            // TODO only subscribe when sth is listening to the event-chain
                            this._subs.push(this.socket.messages$.subscribe(function (cE) {
                                return _this.$emit(cE);
                            }));

                        case 27:
                        case 'end':
                            return _context2.stop();
                    }
                }
            }, _callee2, this, [[4, 10], [13, 18]]);
        }));

        function prepare() {
            return _ref2.apply(this, arguments);
        }

        return prepare;
    }();

    /**
     * spawns a new pouch-instance
     * @param {string} collectionName
     * @param {string} schemaVersion
     * @param {Object} [pouchSettings={}] pouchSettings
     * @type {Object}
     */
    RxDatabase.prototype._spawnPouchDB = function _spawnPouchDB(collectionName, schemaVersion) {
        var pouchSettings = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};

        return _spawnPouchDB2(this.name, this.adapter, collectionName, schemaVersion, pouchSettings, this.pouchSettings);
    };

    RxDatabase.prototype.waitForLeadership = function () {
        var _ref3 = _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee3() {
            return _regeneratorRuntime.wrap(function _callee3$(_context3) {
                while (1) {
                    switch (_context3.prev = _context3.next) {
                        case 0:
                            if (this.multiInstance) {
                                _context3.next = 2;
                                break;
                            }

                            return _context3.abrupt('return', true);

                        case 2:
                            return _context3.abrupt('return', this.leaderElector.waitForLeadership());

                        case 3:
                        case 'end':
                            return _context3.stop();
                    }
                }
            }, _callee3, this);
        }));

        function waitForLeadership() {
            return _ref3.apply(this, arguments);
        }

        return waitForLeadership;
    }();

    /**
     * writes the changeEvent to the socket
     * @param  {RxChangeEvent} changeEvent
     * @return {Promise<boolean>}
     */


    RxDatabase.prototype.writeToSocket = function writeToSocket(changeEvent) {
        if (this.multiInstance && !changeEvent.isIntern() && this.socket) {
            return this.socket.write(changeEvent).then(function () {
                return true;
            });
        } else return Promise.resolve(false);
    };

    /**
     * throw a new event into the event-cicle
     */


    RxDatabase.prototype.$emit = function $emit(changeEvent) {
        if (!changeEvent) return;

        // throw in own cycle
        this.subject.next(changeEvent);

        // write to socket if event was created by self
        if (changeEvent.data.it === this.token) this.writeToSocket(changeEvent);
    };

    /**
     * @return {Observable} observable
     */


    /**
     * returns the primary for a given collection-data
     * used in the internal pouchdb-instances
     * @param {string} name
     * @param {RxSchema} schema
     */
    RxDatabase.prototype._collectionNamePrimary = function _collectionNamePrimary(name, schema) {
        return name + '-' + schema.version;
    };

    /**
     * removes the collection-doc from this._collectionsPouch
     * @return {Promise}
     */


    RxDatabase.prototype.removeCollectionDoc = function removeCollectionDoc(name, schema) {
        var _this2 = this;

        var docId = this._collectionNamePrimary(name, schema);
        return this._collectionsPouch.get(docId).then(function (doc) {
            return _this2.lockedRun(function () {
                return _this2._collectionsPouch.remove(doc);
            });
        });
    };

    /**
     * removes all internal docs of a given collection
     * @param  {string}  collectionName
     * @return {Promise<string[]>} resolves all known collection-versions
     */


    RxDatabase.prototype._removeAllOfCollection = function () {
        var _ref4 = _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee4(collectionName) {
            var _this3 = this;

            var data, relevantDocs;
            return _regeneratorRuntime.wrap(function _callee4$(_context4) {
                while (1) {
                    switch (_context4.prev = _context4.next) {
                        case 0:
                            _context4.next = 2;
                            return this.lockedRun(function () {
                                return _this3._collectionsPouch.allDocs({
                                    include_docs: true
                                });
                            });

                        case 2:
                            data = _context4.sent;
                            relevantDocs = data.rows.map(function (row) {
                                return row.doc;
                            }).filter(function (doc) {
                                var name = doc._id.split('-')[0];
                                return name === collectionName;
                            });
                            _context4.next = 6;
                            return Promise.all(relevantDocs.map(function (doc) {
                                return _this3.lockedRun(function () {
                                    return _this3._collectionsPouch.remove(doc);
                                });
                            }));

                        case 6:
                            return _context4.abrupt('return', relevantDocs.map(function (doc) {
                                return doc.version;
                            }));

                        case 7:
                        case 'end':
                            return _context4.stop();
                    }
                }
            }, _callee4, this);
        }));

        function _removeAllOfCollection(_x2) {
            return _ref4.apply(this, arguments);
        }

        return _removeAllOfCollection;
    }();

    /**
     * create or fetch a collection
     * @param {{name: string, schema: Object, pouchSettings = {}, migrationStrategies = {}}} args
     * @return {Collection}
     */


    RxDatabase.prototype.collection = function () {
        var _ref5 = _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee5(args) {
            var _this4 = this;

            var internalPrimary, schemaHash, collectionDoc, pouch, oneDoc, collection, cEvent;
            return _regeneratorRuntime.wrap(function _callee5$(_context5) {
                while (1) {
                    switch (_context5.prev = _context5.next) {
                        case 0:
                            if (!(typeof args === 'string')) {
                                _context5.next = 2;
                                break;
                            }

                            return _context5.abrupt('return', this.collections[args]);

                        case 2:

                            args.database = this;

                            runPluginHooks('preCreateRxCollection', args);

                            if (!(args.name.charAt(0) === '_')) {
                                _context5.next = 6;
                                break;
                            }

                            throw RxError.newRxError('DB2', {
                                name: args.name
                            });

                        case 6:
                            if (!this.collections[args.name]) {
                                _context5.next = 8;
                                break;
                            }

                            throw RxError.newRxError('DB3', {
                                name: args.name
                            });

                        case 8:
                            if (args.schema) {
                                _context5.next = 10;
                                break;
                            }

                            throw RxError.newRxError('DB4', {
                                name: args.name,
                                args: args
                            });

                        case 10:

                            if (!RxSchema.isInstanceOf(args.schema)) args.schema = RxSchema.create(args.schema);

                            internalPrimary = this._collectionNamePrimary(args.name, args.schema);

                            // check unallowed collection-names

                            if (!properties().includes(args.name)) {
                                _context5.next = 14;
                                break;
                            }

                            throw RxError.newRxError('DB5', {
                                name: args.name
                            });

                        case 14:

                            // check schemaHash
                            schemaHash = args.schema.hash;
                            collectionDoc = null;
                            _context5.prev = 16;
                            _context5.next = 19;
                            return this.lockedRun(function () {
                                return _this4._collectionsPouch.get(internalPrimary);
                            });

                        case 19:
                            collectionDoc = _context5.sent;
                            _context5.next = 24;
                            break;

                        case 22:
                            _context5.prev = 22;
                            _context5.t0 = _context5['catch'](16);

                        case 24:
                            if (!(collectionDoc && collectionDoc.schemaHash !== schemaHash)) {
                                _context5.next = 31;
                                break;
                            }

                            // collection already exists with different schema, check if it has documents
                            pouch = this._spawnPouchDB(args.name, args.schema.version, args.pouchSettings);
                            _context5.next = 28;
                            return pouch.find({
                                selector: {
                                    _id: {}
                                },
                                limit: 1
                            });

                        case 28:
                            oneDoc = _context5.sent;

                            if (!(oneDoc.docs.length !== 0)) {
                                _context5.next = 31;
                                break;
                            }

                            throw RxError.newRxError('DB6', {
                                name: args.name,
                                previousSchemaHash: collectionDoc.schemaHash,
                                schemaHash: schemaHash
                            });

                        case 31:
                            _context5.next = 33;
                            return RxCollection.create(args);

                        case 33:
                            collection = _context5.sent;

                            if (!(Object.keys(collection.schema.encryptedPaths).length > 0 && !this.password)) {
                                _context5.next = 36;
                                break;
                            }

                            throw RxError.newRxError('DB7', {
                                name: args.name
                            });

                        case 36:
                            if (collectionDoc) {
                                _context5.next = 44;
                                break;
                            }

                            _context5.prev = 37;
                            _context5.next = 40;
                            return this.lockedRun(function () {
                                return _this4._collectionsPouch.put({
                                    _id: internalPrimary,
                                    schemaHash: schemaHash,
                                    schema: collection.schema.normalized,
                                    version: collection.schema.version
                                });
                            });

                        case 40:
                            _context5.next = 44;
                            break;

                        case 42:
                            _context5.prev = 42;
                            _context5.t1 = _context5['catch'](37);

                        case 44:
                            cEvent = RxChangeEvent.create('RxDatabase.collection', this);

                            cEvent.data.v = collection.name;
                            cEvent.data.col = '_collections';
                            this.$emit(cEvent);

                            this.collections[args.name] = collection;
                            this.__defineGetter__(args.name, function () {
                                return _this4.collections[args.name];
                            });

                            return _context5.abrupt('return', collection);

                        case 51:
                        case 'end':
                            return _context5.stop();
                    }
                }
            }, _callee5, this, [[16, 22], [37, 42]]);
        }));

        function collection(_x3) {
            return _ref5.apply(this, arguments);
        }

        return collection;
    }();

    /**
     * delete all data of the collection and its previous versions
     * @param  {string}  collectionName
     * @return {Promise}
     */


    RxDatabase.prototype.removeCollection = function () {
        var _ref6 = _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee6(collectionName) {
            var _this5 = this;

            var knownVersions, pouches;
            return _regeneratorRuntime.wrap(function _callee6$(_context6) {
                while (1) {
                    switch (_context6.prev = _context6.next) {
                        case 0:
                            if (!this.collections[collectionName]) {
                                _context6.next = 3;
                                break;
                            }

                            _context6.next = 3;
                            return this.collections[collectionName].destroy();

                        case 3:
                            _context6.next = 5;
                            return this._removeAllOfCollection(collectionName);

                        case 5:
                            knownVersions = _context6.sent;

                            // get all relevant pouchdb-instances
                            pouches = knownVersions.map(function (v) {
                                return _this5._spawnPouchDB(collectionName, v);
                            });

                            // remove documents

                            return _context6.abrupt('return', Promise.all(pouches.map(function (pouch) {
                                return _this5.lockedRun(function () {
                                    return pouch.destroy();
                                });
                            })));

                        case 8:
                        case 'end':
                            return _context6.stop();
                    }
                }
            }, _callee6, this);
        }));

        function removeCollection(_x4) {
            return _ref6.apply(this, arguments);
        }

        return removeCollection;
    }();

    /**
     * runs the given function between idleQueue-locking
     * @return {any}
     */


    RxDatabase.prototype.lockedRun = function lockedRun(fun) {
        return this.idleQueue.wrapCall(fun);
    };

    RxDatabase.prototype.requestIdlePromise = function requestIdlePromise() {
        return this.idleQueue.requestIdlePromise();
    };

    /**
     * export to json
     * @param {boolean} decrypted
     * @param {?string[]} collections array with collectionNames or null if all
     */


    RxDatabase.prototype.dump = function dump() {
        throw RxError.pluginMissing('json-dump');
    };

    /**
     * import json
     * @param {Object} dump
     */


    RxDatabase.prototype.importDump = function importDump() {
        throw RxError.pluginMissing('json-dump');
    };

    /**
     * spawn server
     */


    RxDatabase.prototype.server = function server() {
        throw RxError.pluginMissing('server');
    };

    /**
     * destroys the database-instance and all collections
     * @return {Promise}
     */


    RxDatabase.prototype.destroy = function () {
        var _ref7 = _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee7() {
            var _this6 = this;

            return _regeneratorRuntime.wrap(function _callee7$(_context7) {
                while (1) {
                    switch (_context7.prev = _context7.next) {
                        case 0:
                            if (!this.destroyed) {
                                _context7.next = 2;
                                break;
                            }

                            return _context7.abrupt('return');

                        case 2:
                            runPluginHooks('preDestroyRxDatabase', this);
                            DB_COUNT--;
                            this.destroyed = true;
                            _context7.t0 = this.socket;

                            if (!_context7.t0) {
                                _context7.next = 9;
                                break;
                            }

                            _context7.next = 9;
                            return this.socket.destroy();

                        case 9:
                            if (!this._leaderElector) {
                                _context7.next = 12;
                                break;
                            }

                            _context7.next = 12;
                            return this._leaderElector.destroy();

                        case 12:
                            this._subs.map(function (sub) {
                                return sub.unsubscribe();
                            });

                            // destroy all collections
                            _context7.next = 15;
                            return Promise.all(Object.keys(this.collections).map(function (key) {
                                return _this6.collections[key];
                            }).map(function (col) {
                                return col.destroy();
                            }));

                        case 15:

                            // remove combination from USED_COMBINATIONS-map
                            _removeUsedCombination(this.name, this.adapter);

                        case 16:
                        case 'end':
                            return _context7.stop();
                    }
                }
            }, _callee7, this);
        }));

        function destroy() {
            return _ref7.apply(this, arguments);
        }

        return destroy;
    }();

    /**
     * deletes the database and its stored data
     * @return {Promise}
     */


    RxDatabase.prototype.remove = function remove() {
        var _this7 = this;

        return this.destroy().then(function () {
            return removeDatabase(_this7.name, _this7.adapter);
        });
    };

    _createClass(RxDatabase, [{
        key: '_adminPouch',
        get: function get() {
            if (!this.__adminPouch) this.__adminPouch = _internalAdminPouch(this.name, this.adapter, this.pouchSettings);
            return this.__adminPouch;
        }
    }, {
        key: '_collectionsPouch',
        get: function get() {
            if (!this.__collectionsPouch) this.__collectionsPouch = _internalCollectionsPouch(this.name, this.adapter, this.pouchSettings);
            return this.__collectionsPouch;
        }
    }, {
        key: 'leaderElector',
        get: function get() {
            if (!this._leaderElector) this._leaderElector = overwritable.createLeaderElector(this);
            return this._leaderElector;
        }
    }, {
        key: 'isLeader',
        get: function get() {
            if (!this.multiInstance) return true;
            return this.leaderElector.isLeader;
        }
    }, {
        key: '$',
        get: function get() {
            return this.observable$;
        }
    }]);

    return RxDatabase;
}();

/**
 * returns all possible properties of a RxDatabase-instance
 * @return {string[]} property-names
 */
var _properties = null;
export function properties() {
    if (!_properties) {
        var pseudoInstance = new RxDatabase();
        var ownProperties = Object.getOwnPropertyNames(pseudoInstance);
        var prototypeProperties = Object.getOwnPropertyNames(Object.getPrototypeOf(pseudoInstance));
        _properties = [].concat(ownProperties, prototypeProperties);
    }
    return _properties;
}

/**
 * checks if an instance with same name and adapter already exists
 * @param       {string}  name
 * @param       {any}  adapter
 * @throws {RxError} if used
 */
function _isNameAdapterUsed(name, adapter) {
    if (!USED_COMBINATIONS[name]) return false;

    var used = false;
    USED_COMBINATIONS[name].forEach(function (ad) {
        if (ad === adapter) used = true;
    });
    if (used) {
        throw RxError.newRxError('DB8', {
            name: name,
            adapter: adapter,
            link: 'https://pubkey.github.io/rxdb/rx-database.html#ignoreduplicate'
        });
    }
}

function _removeUsedCombination(name, adapter) {
    if (!USED_COMBINATIONS[name]) return;

    var index = USED_COMBINATIONS[name].indexOf(adapter);
    USED_COMBINATIONS[name].splice(index, 1);
}

export var create = function () {
    var _ref8 = _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee8(_ref9) {
        var name = _ref9.name,
            adapter = _ref9.adapter,
            password = _ref9.password,
            _ref9$multiInstance = _ref9.multiInstance,
            multiInstance = _ref9$multiInstance === undefined ? true : _ref9$multiInstance,
            _ref9$ignoreDuplicate = _ref9.ignoreDuplicate,
            ignoreDuplicate = _ref9$ignoreDuplicate === undefined ? false : _ref9$ignoreDuplicate,
            _ref9$options = _ref9.options,
            options = _ref9$options === undefined ? {} : _ref9$options,
            _ref9$pouchSettings = _ref9.pouchSettings,
            pouchSettings = _ref9$pouchSettings === undefined ? {} : _ref9$pouchSettings;
        var db;
        return _regeneratorRuntime.wrap(function _callee8$(_context8) {
            while (1) {
                switch (_context8.prev = _context8.next) {
                    case 0:
                        util.validateCouchDBString(name);

                        // check if pouchdb-adapter

                        if (!(typeof adapter === 'string')) {
                            _context8.next = 6;
                            break;
                        }

                        if (!(!PouchDB.adapters || !PouchDB.adapters[adapter])) {
                            _context8.next = 4;
                            break;
                        }

                        throw RxError.newRxError('DB9', {
                            adapter: adapter
                        });

                    case 4:
                        _context8.next = 9;
                        break;

                    case 6:
                        util.isLevelDown(adapter);

                        if (!(!PouchDB.adapters || !PouchDB.adapters.leveldb)) {
                            _context8.next = 9;
                            break;
                        }

                        throw RxError.newRxError('DB10', {
                            adapter: adapter
                        });

                    case 9:

                        if (password) overwritable.validatePassword(password);

                        // check if combination already used
                        if (!ignoreDuplicate) _isNameAdapterUsed(name, adapter);

                        // add to used_map
                        if (!USED_COMBINATIONS[name]) USED_COMBINATIONS[name] = [];
                        USED_COMBINATIONS[name].push(adapter);

                        db = new RxDatabase(name, adapter, password, multiInstance, options, pouchSettings);
                        _context8.next = 16;
                        return db.prepare();

                    case 16:

                        runPluginHooks('createRxDatabase', db);
                        return _context8.abrupt('return', db);

                    case 18:
                    case 'end':
                        return _context8.stop();
                }
            }
        }, _callee8, this);
    }));

    return function create(_x5) {
        return _ref8.apply(this, arguments);
    };
}();

function _spawnPouchDB2(dbName, adapter, collectionName, schemaVersion) {
    var pouchSettings = arguments.length > 4 && arguments[4] !== undefined ? arguments[4] : {};
    var pouchSettingsFromRxDatabaseCreator = arguments.length > 5 && arguments[5] !== undefined ? arguments[5] : {};

    var pouchLocation = dbName + '-rxdb-' + schemaVersion + '-' + collectionName;
    var pouchDbParameters = {
        location: pouchLocation,
        adapter: util.adapterObject(adapter),
        settings: pouchSettings
    };
    var pouchDBOptions = Object.assign({}, pouchDbParameters.adapter, pouchSettingsFromRxDatabaseCreator);
    runPluginHooks('preCreatePouchDb', pouchDbParameters);
    return new PouchDB(pouchDbParameters.location, pouchDBOptions, pouchDbParameters.settings);
}

function _internalAdminPouch(name, adapter) {
    var pouchSettingsFromRxDatabaseCreator = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};

    return _spawnPouchDB2(name, adapter, '_admin', 0, {
        auto_compaction: false, // no compaction because this only stores local documents
        revs_limit: 1
    }, pouchSettingsFromRxDatabaseCreator);
}

function _internalCollectionsPouch(name, adapter) {
    var pouchSettingsFromRxDatabaseCreator = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};

    return _spawnPouchDB2(name, adapter, '_collections', 0, {
        auto_compaction: false, // no compaction because this only stores local documents
        revs_limit: 1
    }, pouchSettingsFromRxDatabaseCreator);
}

export var removeDatabase = function () {
    var _ref10 = _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee9(databaseName, adapter) {
        var adminPouch, socketPouch, collectionsPouch, collectionsData;
        return _regeneratorRuntime.wrap(function _callee9$(_context9) {
            while (1) {
                switch (_context9.prev = _context9.next) {
                    case 0:
                        adminPouch = _internalAdminPouch(databaseName, adapter);
                        socketPouch = _spawnPouchDB2(databaseName, adapter, '_socket', 0);
                        collectionsPouch = _internalCollectionsPouch(databaseName, adapter);
                        _context9.next = 5;
                        return collectionsPouch.allDocs({
                            include_docs: true
                        });

                    case 5:
                        collectionsData = _context9.sent;


                        // remove collections
                        Promise.all(collectionsData.rows.map(function (colDoc) {
                            return colDoc.id;
                        }).map(function (id) {
                            var split = id.split('-');
                            var name = split[0];
                            var version = parseInt(split[1], 10);
                            var pouch = _spawnPouchDB2(databaseName, adapter, name, version);
                            return pouch.destroy();
                        }));

                        // remove internals
                        _context9.next = 9;
                        return Promise.all([collectionsPouch.destroy(), adminPouch.destroy(), socketPouch.destroy()]);

                    case 9:
                    case 'end':
                        return _context9.stop();
                }
            }
        }, _callee9, this);
    }));

    return function removeDatabase(_x10, _x11) {
        return _ref10.apply(this, arguments);
    };
}();

/**
 * check is the given adapter can be used
 */
export var checkAdapter = function () {
    var _ref11 = _asyncToGenerator( /*#__PURE__*/_regeneratorRuntime.mark(function _callee10(adapter) {
        return _regeneratorRuntime.wrap(function _callee10$(_context10) {
            while (1) {
                switch (_context10.prev = _context10.next) {
                    case 0:
                        _context10.next = 2;
                        return overwritable.checkAdapter(adapter);

                    case 2:
                        return _context10.abrupt('return', _context10.sent);

                    case 3:
                    case 'end':
                        return _context10.stop();
                }
            }
        }, _callee10, this);
    }));

    return function checkAdapter(_x12) {
        return _ref11.apply(this, arguments);
    };
}();;

export function isInstanceOf(obj) {
    return obj instanceof RxDatabase;
}

export function dbCount() {
    return DB_COUNT;
}

// TODO is this needed?
export { RxSchema };

export default {
    create: create,
    removeDatabase: removeDatabase,
    checkAdapter: checkAdapter,
    isInstanceOf: isInstanceOf,
    RxDatabase: RxDatabase,
    RxSchema: RxSchema,
    dbCount: dbCount
};