import { Observable } from 'rxjs';
import {
    RxCollectionCreator,
    RxCollection
} from './rx-collection';

import {
    RxLocalDocument
} from './rx-document';
import {
    RxChangeEventInsert,
    RxChangeEventUpdate,
    RxChangeEventRemove,
    RxChangeEventCollection
} from './rx-change-event';

import {
    PouchSettings
} from "./pouch";

export interface RxDatabaseCreator {
    name: string;
    adapter: any;
    password?: string;
    multiInstance?: boolean;
    ignoreDuplicate?: boolean;
    options?: any;
    pouchSettings?: PouchSettings;
}

export declare class RxDatabase {
    readonly name: string;
    readonly token: string;
    readonly multiInstance: boolean;
    readonly password: string;
    readonly collections: any;
    options?: any;
    pouchSettings?: PouchSettings;

    readonly $: Observable<RxChangeEventInsert<any> | RxChangeEventUpdate<any> | RxChangeEventRemove<any> | RxChangeEventCollection>;

    collection(args: RxCollectionCreator): Promise<RxCollection<any>>;
    destroy(): Promise<boolean>;
    dump(): Promise<any>;
    importDump(json: any): Promise<any>;
    remove(): Promise<any>;

    readonly isLeader: boolean;

    insertLocal(id: string, data: any): Promise<RxLocalDocument<RxDatabase>>;
    upsertLocal(id: string, data: any): Promise<RxLocalDocument<RxDatabase>>;
    getLocal(id: string): Promise<RxLocalDocument<RxDatabase>>;

    /**
     * returns a promise which resolves when the instance becomes leader
     * @return {Promise<boolean>}
     */
    waitForLeadership(): Promise<boolean>;

    /**
     * removes all internal collection-info
     * only use this if you have to upgrade from a major rxdb-version
     * do NEVER use this to change the schema of a collection
     */
    dangerousRemoveCollectionInfo(): Promise<void>;
}
